"""Project file loader."""
from __future__ import annotations

import re
from xml.etree import ElementTree

from xknxproject.models import (
    ChannelNode,
    ComObjectInstanceRef,
    DeviceInstance,
    GroupAddressStyle,
    KNXMasterData,
    ModuleInstance,
    ModuleInstanceArgument,
    SpaceType,
    XMLArea,
    XMLFunction,
    XMLGroupAddress,
    XMLGroupAddressRef,
    XMLGroupRange,
    XMLLine,
    XMLProjectInformation,
    XMLSpace,
)
from xknxproject.util import get_dpt_type, parse_dpt_types, parse_xml_flag
from xknxproject.zip import KNXProjContents


class ProjectLoader:
    """Load Project file."""

    @staticmethod
    def load(
        knx_proj_contents: KNXProjContents,
        knx_master_data: KNXMasterData,
    ) -> tuple[
        list[XMLGroupAddress],
        list[XMLGroupRange],
        list[XMLArea],
        list[DeviceInstance],
        list[XMLSpace],
        XMLProjectInformation,
        list[XMLFunction],
    ]:
        """Load topology mappings."""
        areas: list[XMLArea] = []
        devices: list[DeviceInstance] = []
        group_address_list: list[XMLGroupAddress] = []
        group_range_list: list[XMLGroupRange] = []
        spaces: list[XMLSpace] = []
        functions: list[XMLFunction] = []

        with knx_proj_contents.open_project_meta() as project_file:
            tree = ElementTree.parse(project_file)
            project_info = load_project_info(tree)

        with knx_proj_contents.open_project_0() as project_0_file:
            tree = ElementTree.parse(project_0_file)
            for ga_element in tree.findall(
                # `//` to ignore <GroupRange> tags to support different GA level formats
                "{*}Project/{*}Installations/{*}Installation/{*}GroupAddresses//{*}GroupAddress"
            ):
                group_address_list.append(
                    _GroupAddressLoader.load(
                        group_address_element=ga_element,
                        group_address_style=project_info.group_address_style,
                    ),
                )
            for ga_range_l1 in tree.findall(
                "{*}Project/{*}Installations/{*}Installation/{*}GroupAddresses/{*}GroupRanges/{*}GroupRange"
            ):
                group_range_list.append(
                    _GroupAddressRangeLoader.load(
                        ga_range_l1, project_info.group_address_style
                    )
                )
            topology_loader = _TopologyLoader(knx_proj_contents)
            for topology_element in tree.findall(
                "{*}Project/{*}Installations/{*}Installation/{*}Topology"
            ):
                areas.extend(topology_loader.load(topology_element=topology_element))
            for area in areas:
                for line in area.lines:
                    devices.extend(line.devices)

            # ETS4 has a different naming for locations than ETS5/6
            element_name = (
                "Buildings" if knx_proj_contents.is_ets4_project() else "Locations"
            )

            location_loader = _LocationLoader(
                knx_proj_contents,
                knx_master_data,
                devices,
            )
            for location_element in tree.findall(
                f"{{*}}Project/{{*}}Installations/{{*}}Installation/{{*}}{element_name}"
            ):
                spaces.extend(
                    location_loader.load(
                        location_element=location_element, functions=functions
                    )
                )

        for function in functions:
            function.usage_text = (
                knx_master_data.get_function_type_name(function.function_type)
                if function.function_type
                else ""
            )

            for group_address in function.group_addresses:
                group_address.address = next(
                    ga.address
                    for ga in group_address_list
                    if ga.identifier == group_address.ref_id
                )

        return (
            group_address_list,
            group_range_list,
            areas,
            devices,
            spaces,
            project_info,
            functions,
        )


class _GroupAddressLoader:
    """Load GroupAddress info from KNX XML."""

    @staticmethod
    def load(
        group_address_element: ElementTree.Element,
        group_address_style: GroupAddressStyle,
    ) -> XMLGroupAddress:
        """Load GroupAddress mappings."""
        project_uid = group_address_element.get("Puid")
        return XMLGroupAddress(
            name=group_address_element.get("Name", ""),
            identifier=group_address_element.get("Id", ""),
            address=group_address_element.get("Address", ""),
            project_uid=int(project_uid) if project_uid else None,
            description=group_address_element.get("Description", ""),
            dpt=get_dpt_type(group_address_element.get("DatapointType")),
            data_secure_key=group_address_element.get("Key"),
            comment=group_address_element.get("Comment", ""),
            style=group_address_style,
        )


class _GroupAddressRangeLoader:
    """Load information of GroupAddress levels (GroupRange)."""

    @staticmethod
    def load(
        group_range_element: ElementTree.Element, group_address_style: GroupAddressStyle
    ) -> XMLGroupRange:
        """Load GroupRange."""

        def create_xml_group_range(elem: ElementTree.Element) -> XMLGroupRange:
            group_range_elems = elem.findall("./{*}GroupRange")
            group_ranges = [
                create_xml_group_range(range_elem) for range_elem in group_range_elems
            ]

            return XMLGroupRange(
                name=elem.get("Name", ""),
                range_start=int(elem.get("RangeStart")),  # type: ignore[arg-type]
                range_end=int(elem.get("RangeEnd")),  # type: ignore[arg-type]
                group_addresses=[
                    int(e.attrib["Address"]) for e in elem.findall("{*}GroupAddress")
                ],
                group_ranges=group_ranges,
                comment=elem.get("Comment", ""),
                style=group_address_style,
            )

        return create_xml_group_range(group_range_element)


class _TopologyLoader:
    """Load topology from KNX XML."""

    def __init__(self, knx_proj_contents: KNXProjContents) -> None:
        self.__knx_proj_contents = knx_proj_contents

    def load(self, topology_element: ElementTree.Element) -> list[XMLArea]:
        """Load topology mappings."""
        areas: list[XMLArea] = []

        for area in topology_element.findall("{*}Area"):
            areas.append(self._create_area(area))

        return areas

    def _create_area(self, area_element: ElementTree.Element) -> XMLArea:
        """Create an Area."""
        address: int = int(area_element.get("Address", ""))
        name: str = area_element.get("Name", "")
        description: str | None = area_element.get("Description")
        area: XMLArea = XMLArea(address, name, description, [])

        for line_element in area_element:
            area.lines.append(self._create_line(line_element, area))

        return area

    def _create_line(self, line_element: ElementTree.Element, area: XMLArea) -> XMLLine:
        """Create a Line."""
        address: int = int(line_element.get("Address", ""))
        name: str = line_element.get("Name", "")
        description: str | None = line_element.get("Description")
        if (segment := line_element.find("{*}Segment")) is not None:
            #  ETS-6 (21) adds "Segment" tags between "Line" and "DeviceInstance" tags
            medium_type = segment.get("MediumTypeRefId", "")
        else:
            medium_type = line_element.get("MediumTypeRefId", "")
        line: XMLLine = XMLLine(address, description, name, medium_type, [], area)

        for device_element in line_element.findall(".//{*}DeviceInstance"):
            if device := self._create_device(device_element, line):
                line.devices.append(device)

        return line

    def _create_device(
        self, device_element: ElementTree.Element, line: XMLLine
    ) -> DeviceInstance | None:
        """Create device."""
        address: str | None = device_element.get("Address")
        #  devices like power supplies do usually not have an IA.
        if address is None:
            return None

        project_uid = device_element.get("Puid")
        product_ref = device_element.get("ProductRefId", "")
        additional_addresses = [
            add_addr
            for address_elem in device_element.findall(
                "{*}AdditionalAddresses/{*}Address"
            )
            if (add_addr := address_elem.get("Address")) is not None
        ]
        com_obj_inst_refs = [
            com_obj_inst_ref
            for elem in device_element.findall(
                "{*}ComObjectInstanceRefs/{*}ComObjectInstanceRef"
            )
            if (com_obj_inst_ref := self._create_com_object_instance(elem)) is not None
        ]
        module_instances = [
            module_instance
            for mi_elem in device_element.findall(
                "{*}ModuleInstances/{*}ModuleInstance"
            )
            if (module_instance := self._create_module_instance(mi_elem)) is not None
        ]
        channels = [
            ChannelNode(
                ref_id=channel_node_elem.get("RefId"),  # type: ignore[arg-type]
                name=channel_node_elem.get("Text", ""),
            )
            for channel_node_elem in device_element.findall(
                "{*}GroupObjectTree/{*}Nodes/{*}Node[@Type='Channel']"
            )
        ]
        return DeviceInstance(
            identifier=device_element.get("Id", ""),
            address=int(address),
            project_uid=int(project_uid) if project_uid else None,
            name=device_element.get("Name", ""),
            description=device_element.get("Description", ""),
            last_modified=device_element.get("LastModified", ""),
            product_ref=product_ref,
            hardware_program_ref=device_element.get("Hardware2ProgramRefId", ""),
            line=line,
            manufacturer=product_ref.split("_", 1)[0],
            additional_addresses=additional_addresses,
            channels=channels,
            com_object_instance_refs=com_obj_inst_refs,
            module_instances=module_instances,
        )

    @staticmethod
    def __get_links_from_ets4(com_object: ElementTree.Element) -> list[str]:
        # Check if "Connectors" is available. This will always fail for ETS5/6
        if (connectors := com_object.find("{*}Connectors")) is None:
            return []

        # Send GA is the primary GA, Receive GA are additional group addresses
        ga_list = connectors.findall("{*}Send") + connectors.findall("{*}Receive")

        # Remove the project ID from GA
        return [ga.get("GroupAddressRefId", "").split("_")[1] for ga in ga_list]

    @staticmethod
    def __get_links_from_ets5(com_object: ElementTree.Element) -> list[str]:
        # ETS5/6 uses a space-separated string of GA
        links = com_object.get("Links")

        if links is None:
            return []

        return links.split(" ")

    def _create_com_object_instance(
        self,
        com_object: ElementTree.Element,
    ) -> ComObjectInstanceRef | None:
        """Create ComObjectInstanceRef."""

        if self.__knx_proj_contents.is_ets4_project():
            links = self.__get_links_from_ets4(com_object)
        else:
            links = self.__get_links_from_ets5(com_object)

        if not links:
            return None

        return ComObjectInstanceRef(
            identifier=com_object.get("Id"),
            ref_id=com_object.get("RefId"),  # type: ignore[arg-type]
            text=com_object.get("Text"),
            function_text=com_object.get("FunctionText"),
            read_flag=parse_xml_flag(com_object.get("ReadFlag")),
            write_flag=parse_xml_flag(com_object.get("WriteFlag")),
            communication_flag=parse_xml_flag(com_object.get("CommunicationFlag")),
            transmit_flag=parse_xml_flag(com_object.get("TransmitFlag")),
            update_flag=parse_xml_flag(com_object.get("UpdateFlag")),
            read_on_init_flag=parse_xml_flag(com_object.get("ReadOnInitFlag")),
            datapoint_types=parse_dpt_types(com_object.get("DatapointType")),
            description=com_object.get("Description"),
            channel=com_object.get("ChannelId"),
            links=links,
        )

    def _create_module_instance(
        self,
        module_instance_elem: ElementTree.Element,
    ) -> ModuleInstance | None:
        """Create ComObjectInstanceRef."""
        module_arguments = [
            ModuleInstanceArgument(
                ref_id=arg.get("RefId"),  # type: ignore[arg-type]
                value=arg.get("Value"),  # type: ignore[arg-type]
            )
            for arg in module_instance_elem.findall("{*}Arguments/{*}Argument")
        ]
        return ModuleInstance(
            identifier=module_instance_elem.get("Id"),  # type: ignore[arg-type]
            ref_id=module_instance_elem.get("RefId"),  # type: ignore[arg-type]
            arguments=module_arguments,
        )


class _LocationLoader:
    """Load location infos from KNX XML."""

    def __init__(
        self,
        knx_proj_contents: KNXProjContents,
        knx_master_data: KNXMasterData,
        devices: list[DeviceInstance],
    ):
        """Initialize the LocationLoader."""
        self.knx_master_data = knx_master_data
        self._element_name = (
            "BuildingPart" if knx_proj_contents.is_ets4_project() else "Space"
        )
        self.devices: dict[str, str] = {
            device.identifier: device.individual_address for device in devices
        }

    def load(
        self, location_element: ElementTree.Element, functions: list[XMLFunction]
    ) -> list[XMLSpace]:
        """Load Location mappings."""
        return [
            self.parse_space(space, functions)
            for space in location_element.findall(f"{{*}}{self._element_name}")
        ]

    def parse_space(
        self, node: ElementTree.Element, functions: list[XMLFunction]
    ) -> XMLSpace:
        """Parse a space from the document."""
        usage_id = node.get("Usage")
        usage_text = (
            self.knx_master_data.get_space_usage_name(usage_id) if usage_id else ""
        )
        project_uid = node.get("Puid")
        space: XMLSpace = XMLSpace(
            identifier=node.get("Id"),  # type: ignore[arg-type]
            name=node.get("Name"),  # type: ignore[arg-type]
            space_type=SpaceType(node.get("Type")),
            usage_id=usage_id,
            usage_text=usage_text,
            number=node.get("Number", ""),
            description=node.get("Description", ""),
            project_uid=int(project_uid) if project_uid else None,
            spaces=[],
            devices=[],
            functions=[],
        )

        for sub_node in node:
            if sub_node.tag.endswith(self._element_name):
                # recursively call parse space since this can be nested for an unbound time in the XSD
                space.spaces.append(self.parse_space(sub_node, functions))
            elif sub_node.tag.endswith("DeviceInstanceRef"):
                if individual_address := self.devices.get(sub_node.get("RefId", "")):
                    space.devices.append(individual_address)
            elif sub_node.tag.endswith("Function"):
                function = self.parse_functions(sub_node)
                function.space_id = space.identifier
                functions.append(function)
                space.functions.append(function.identifier)

        return space

    def parse_functions(self, node: ElementTree.Element) -> XMLFunction:
        """Parse a functions from the document."""
        identifier = node.get("Id", "").split("_")[1]
        project_uid = node.get("Puid")
        function_type = node.get("Type", "")

        functions: XMLFunction = XMLFunction(
            identifier=identifier,
            name=node.get("Name"),  # type: ignore[arg-type]
            function_type=function_type,
            project_uid=int(project_uid) if project_uid else None,
            group_addresses=[],
            usage_text="",
            space_id="",
        )

        for sub_node in node:
            if sub_node.tag.endswith("GroupAddressRef"):
                project_uid = sub_node.get("Puid")
                ref_id = sub_node.get("RefId", "").split("_")[1]

                group_address_ref: XMLGroupAddressRef = XMLGroupAddressRef(
                    ref_id=ref_id,
                    identifier=sub_node.get("Id"),  # type: ignore[arg-type]
                    name=sub_node.get("Name"),  # type: ignore[arg-type]
                    role=sub_node.get("Role", ""),
                    project_uid=int(project_uid) if project_uid else None,
                    address="",
                )
                functions.group_addresses.append(group_address_ref)

        return functions


def load_project_info(tree: ElementTree.ElementTree) -> XMLProjectInformation:
    """Load project information."""
    knx_root = tree.getroot()
    _namespace_match = re.match(r"{.+\/project\/(.+)}", knx_root.tag)
    schema_version = _namespace_match.group(1) if _namespace_match else ""
    created_by = knx_root.get("CreatedBy", "")
    tool_version = knx_root.get("ToolVersion", "")

    try:
        project_node: ElementTree.Element = tree.find("{*}Project")  # type: ignore[assignment]
        identifier = project_node.get("Id", "")
        info_node: ElementTree.Element = project_node.find("{*}ProjectInformation")  # type: ignore[assignment]
        return XMLProjectInformation(
            project_id=identifier,
            name=info_node.get("Name", ""),
            last_modified=info_node.get("LastModified"),
            group_address_style=GroupAddressStyle(info_node.get("GroupAddressStyle")),
            guid=info_node.get("Guid"),  # type: ignore[arg-type]
            created_by=created_by,
            schema_version=schema_version,
            tool_version=tool_version,
        )
    except AttributeError:
        # Project or ProjectInformation tag not found
        return XMLProjectInformation(
            created_by=created_by,
            schema_version=schema_version,
            tool_version=tool_version,
        )



"""Contains static data from the project files that is not really important and doesn't change often."""
from enum import Enum


class SpaceType(Enum):
    """Supported space types according to Specs from XSD."""

    BUILDING = "Building"
    BUILDING_PART = "BuildingPart"
    FLOOR = "Floor"
    ROOM = "Room"
    DISTRIBUTION_BOARD = "DistributionBoard"
    STAIRWAY = "Stairway"
    CORRIDOR = "Corridor"
    AREA = "Area"
    GROUND = "Ground"
    SEGMENT = "Segment"


MEDIUM_TYPES: dict[str, str] = {
    "MT-0": "Twisted Pair (TP)",
    "MT-1": "Powerline (PL)",
    "MT-2": "KNX RF (RF)",
    "MT-5": "KNXnet/IP (IP)",
}


class GroupAddressStyle(Enum):
    """Supported GroupAddress style identifiers."""

    TWOLEVEL = "TwoLevel"
    THREELEVEL = "ThreeLevel"
    FREE = "Free"



"""KNX Master Data Loader."""
from __future__ import annotations

import logging
from xml.etree import ElementTree
from zipfile import Path

from xknxproject.const import ETS4_PRODUCT_LANGUAGES
from xknxproject.models import KNXMasterData, TranslationsType
from xknxproject.zip import KNXProjContents

_LOGGER = logging.getLogger("xknxproject.log")


class KNXMasterLoader:
    """Load KNX master XML."""

    @staticmethod
    def load(
        knx_proj_contents: KNXProjContents,
        knx_master_file: Path,
        language: str | None,
    ) -> tuple[KNXMasterData, str | None]:
        """Load KNX master data. Returns KNXMasterData and the found language code."""
        manufacturer_mapping: dict[str, str] = {}
        space_usage_mapping: dict[str, str] = {}
        product_languages: list[str] = []  # eg. "en-US", "de-DE", "fr-FR"
        language_code: str | None = None
        translations: TranslationsType = {}
        function_type_mapping: dict[str, str] = {}

        with knx_master_file.open(mode="rb") as master_xml:
            tree = ElementTree.parse(master_xml)
            for manufacturer in tree.findall(".//{*}Manufacturers/{*}Manufacturer"):
                identifier = manufacturer.get("Id", "")
                manufacturer_mapping[identifier] = manufacturer.get("Name", "")

            if knx_proj_contents.is_ets4_project():
                # No SpaceUsage in ETS4, neither ProductLanguages. Therefore, we use an
                # hardcoded list of common product languages
                product_languages = ETS4_PRODUCT_LANGUAGES
            else:
                for space_usage_node in tree.findall(".//{*}SpaceUsages/{*}SpaceUsage"):
                    identifier = space_usage_node.get("Id", "")
                    space_usage_mapping[identifier] = space_usage_node.get("Text", "")

                for language_node in tree.findall(".//{*}ProductLanguages/{*}Language"):
                    product_languages.append(language_node.get("Identifier", ""))

                for function_type_node in tree.findall(
                    ".//{*}FunctionTypes/{*}FunctionType"
                ):
                    identifier = function_type_node.get("Id", "")
                    function_type_mapping[identifier] = function_type_node.get(
                        "Text", ""
                    )

            if language is not None:
                language_code = KNXMasterLoader.get_language_code(
                    language, product_languages
                )

            if language_code:
                for translation_element in tree.findall(
                    ".//{*}Languages"
                    f"/{{*}}Language[@Identifier='{language_code}']"
                    "/{*}TranslationUnit/{*}TranslationElement"
                ):
                    _ref_id = translation_element.get("RefId", "")
                    translations[_ref_id] = {
                        attr: text
                        for item in translation_element.findall("{*}Translation")
                        if (attr := item.get("AttributeName")) is not None
                        and (text := item.get("Text")) is not None
                    }

        return (
            KNXMasterData(
                function_type_names=function_type_mapping,
                manufacturer_names=manufacturer_mapping,
                space_usage_mapping=space_usage_mapping,
                translations=translations,
            ),
            language_code,
        )

    @staticmethod
    def get_language_code(language: str, product_languages: list[str]) -> str | None:
        """Infer language code from product languages."""
        if language in product_languages:
            return language

        for language_code, country_code in [
            _lan.split("-", maxsplit=1) for _lan in product_languages
        ]:
            if language[:2].lower() == language_code:
                used_language = f"{language_code}-{country_code}"
                _LOGGER.info(
                    'Using language code "%s" for "%s"', used_language, language
                )
                return used_language

        _LOGGER.warning("No matching language code found for %s", language)
        return None



"""Package for exception handling."""
from .exceptions import (
    InvalidPasswordException,
    ProjectNotFoundException,
    UnexpectedFileContent,
    XknxProjectException,
)

__all__ = [
    "XknxProjectException",
    "InvalidPasswordException",
    "ProjectNotFoundException",
    "UnexpectedFileContent",
]



"""Class to read KNXProj ZIP files."""
from __future__ import annotations

import base64
from collections.abc import Iterator
from contextlib import contextmanager
import hashlib
import logging
from pathlib import Path
import re
from typing import IO
from zipfile import Path as ZipPath, ZipFile, ZipInfo

import pyzipper

from xknxproject.const import ETS_4_2_SCHEMA_VERSION, ETS_6_SCHEMA_VERSION
from xknxproject.exceptions import (
    InvalidPasswordException,
    ProjectNotFoundException,
    UnexpectedFileContent,
)

_LOGGER = logging.getLogger("xknxproject.log")


class KNXProjContents:
    """Class for holding the contents of a KNXProj file."""

    def __init__(
        self,
        root_zip: ZipFile,
        project_archive: ZipFile,
        project_relative_path: str,
        xml_namespace: str,
    ):
        """Initialize a KNXProjContents."""
        self._project_archive = project_archive
        self._project_relative_path = project_relative_path
        self.root = root_zip
        self.root_path = ZipPath(root_zip)
        self.xml_namespace = xml_namespace
        self.schema_version = _get_schema_version(xml_namespace)

    def is_ets4_project(self) -> bool:
        """Check if the project is an ETS4 project."""
        return self.schema_version <= ETS_4_2_SCHEMA_VERSION

    def open_project_0(self) -> IO[bytes]:
        """Open the project 0.xml file."""
        return self._project_archive.open(
            f"{self._project_relative_path}0.xml",
            mode="r",
        )

    def open_project_meta(self) -> IO[bytes]:
        """Open the project.xml file."""
        project_filename = "Project.xml" if self.is_ets4_project() else "project.xml"
        return self._project_archive.open(
            f"{self._project_relative_path}{project_filename}",
            mode="r",
        )


@contextmanager
def extract(
    archive_path: Path, password: str | None = None
) -> Iterator[KNXProjContents]:
    """Provide the contents of a KNXProj file."""
    _LOGGER.debug('Opening KNX Project file "%s"', archive_path)
    with ZipFile(archive_path, mode="r") as zip_archive:
        project_id = _get_project_id(zip_archive)
        xml_namespace = _get_xml_namespace(zip_archive)

        password_protected: bool
        try:
            protected_info = zip_archive.getinfo(name=project_id + ".zip")
        except KeyError:
            _LOGGER.debug("Project %s is not password protected", project_id)
            password_protected = False
            # move yield out of except block to clear exception context
        else:
            _LOGGER.debug("Project %s is password protected", project_id)
            password_protected = True

        if not password_protected:
            yield KNXProjContents(
                root_zip=zip_archive,
                project_archive=zip_archive,
                project_relative_path=f"{project_id}/",
                xml_namespace=xml_namespace,
            )
            return
        # Password protected project
        schema_version = _get_schema_version(xml_namespace)
        with _extract_protected_project_file(
            zip_archive, protected_info, password, schema_version
        ) as project_zip:
            # ZipPath is not supported by pyzipper thus we use
            # string name for project_relative_path
            yield KNXProjContents(
                root_zip=zip_archive,
                project_archive=project_zip,
                project_relative_path="",
                xml_namespace=xml_namespace,
            )


def _get_project_id(zip_archive: ZipFile) -> str:
    """Get the project id."""
    for info in zip_archive.infolist():
        if info.filename.startswith("P-") and info.filename.endswith(".signature"):
            return info.filename.removesuffix(".signature")

    raise ProjectNotFoundException("Signature file not found.")


@contextmanager
def _extract_protected_project_file(
    archive_zip: ZipFile, info: ZipInfo, password: str | None, schema_version: int
) -> Iterator[ZipFile]:
    """Unzip a protected ETS5/6 project file."""
    if not password:
        raise InvalidPasswordException("Password required.")

    try:
        project_archive: ZipFile
        # Password protection is different for ETS4/5 and ETS6
        if schema_version < ETS_6_SCHEMA_VERSION:
            project_archive = ZipFile(archive_zip.open(info, mode="r"), mode="r")
            project_archive.setpassword(password.encode("utf-8"))
            yield project_archive
        else:
            project_archive = pyzipper.AESZipFile(
                archive_zip.open(info, mode="r"), mode="r"
            )
            project_archive.setpassword(_generate_ets6_zip_password(password))
            yield project_archive
    except RuntimeError as exception:
        raise InvalidPasswordException("Invalid password.") from exception


def _get_xml_namespace(project_zip: ZipFile) -> str:
    """Get the XML namespace of the project."""
    with project_zip.open("knx_master.xml", mode="r") as master:
        for line_number, line in enumerate(master, start=1):
            if line_number == 2:
                try:
                    namespace_match = re.match(
                        r".+ xmlns=\"(.+?)\"",
                        line.decode(),
                    )
                    namespace = namespace_match.group(1)  # type: ignore[union-attr]
                except (AttributeError, IndexError, UnicodeDecodeError):
                    _LOGGER.error("Could not parse XML namespace from %s", line)
                    raise UnexpectedFileContent("Could not parse XML namespace.")

        _LOGGER.debug("Namespace: %s", namespace)
        return namespace


def _get_schema_version(namespace: str) -> int:
    """Get the schema version of the project."""
    try:
        schema_version = int(namespace.split("/")[-1])
    except ValueError:
        _LOGGER.error("Could not parse schema version from %s", namespace)
        raise UnexpectedFileContent("Could not parse schema version.")

    _LOGGER.debug("Schema version: %s", schema_version)
    return schema_version


def _generate_ets6_zip_password(password: str) -> bytes:
    """Generate ZIP archive password."""

    return base64.b64encode(
        hashlib.pbkdf2_hmac(
            hash_name="sha256",
            password=password.encode("utf-16-le"),
            salt=b"21.project.ets.knx.org",
            iterations=65536,
            dklen=32,
        )
    )



"""Parser logic for ETS XML files."""
from __future__ import annotations

import html
import logging
from operator import attrgetter

from striprtf.striprtf import rtf_to_text

from xknxproject.__version__ import __version__
from xknxproject.loader import (
    ApplicationProgramLoader,
    HardwareLoader,
    KNXMasterLoader,
    ProjectLoader,
)
from xknxproject.models import (
    MEDIUM_TYPES,
    Area,
    Channel,
    CommunicationObject,
    Device,
    DeviceInstance,
    Flags,
    Function,
    GroupAddress,
    GroupAddressRef,
    GroupAddressStyle,
    GroupRange,
    HardwareToPrograms,
    KNXProject,
    Line,
    Product,
    ProjectInfo,
    Space,
    XMLArea,
    XMLFunction,
    XMLGroupAddress,
    XMLGroupAddressRef,
    XMLGroupRange,
    XMLProjectInformation,
    XMLSpace,
)
from xknxproject.zip.extractor import KNXProjContents

_LOGGER = logging.getLogger("xknxproject.log")


def _convert_group_address_ref(
    group_address_ref: XMLGroupAddressRef,
) -> GroupAddressRef:
    """Convert group address ref to the final output format."""
    return GroupAddressRef(
        address=group_address_ref.address,
        name=group_address_ref.name,
        project_uid=group_address_ref.project_uid,
        role=group_address_ref.role,
    )


def _convert_functions(function: XMLFunction) -> Function:
    """Convert function to the final output format."""

    ga_dict = {}
    for group_address in function.group_addresses:
        ga_dict[group_address.address] = _convert_group_address_ref(group_address)

    return Function(
        function_type=function.function_type,
        group_addresses=ga_dict,
        identifier=function.identifier,
        name=function.name,
        project_uid=function.project_uid,
        space_id=function.space_id,
        usage_text=function.usage_text,
    )


def _recursive_convert_spaces(spaces: list[XMLSpace]) -> dict[str, Space]:
    """Convert spaces to the final output format."""
    return {
        space.name: Space(
            type=space.space_type.value,
            identifier=space.identifier,
            name=space.name,
            usage_id=space.usage_id,
            usage_text=space.usage_text,
            number=space.number,
            description=space.description,
            project_uid=space.project_uid,
            devices=space.devices,
            spaces=_recursive_convert_spaces(space.spaces),
            functions=space.functions,
        )
        for space in spaces
    }


def _recursive_convert_group_range(
    group_ranges: list[XMLGroupRange],
    group_address_style: GroupAddressStyle,
) -> dict[str, GroupRange]:
    """Convert XMLGroupRange into GroupRange."""
    return {
        group_range.str_address(): GroupRange(
            name=group_range.name,
            address_start=group_range.range_start,
            address_end=group_range.range_end,
            group_addresses=[
                XMLGroupAddress.str_address(ga, group_address_style)
                for ga in group_range.group_addresses
            ],
            comment=html.unescape(rtf_to_text(group_range.comment)),
            group_ranges=_recursive_convert_group_range(
                group_range.group_ranges,
                group_address_style,
            ),
        )
        for group_range in group_ranges
    }


class XMLParser:
    """Class that parses XMLs and returns useful information."""

    def __init__(self, knx_proj_contents: KNXProjContents) -> None:
        """Initialize the parser."""
        self.knx_proj_contents = knx_proj_contents
        self.spaces: list[XMLSpace] = []
        self.group_addresses: list[XMLGroupAddress] = []
        self.group_ranges: list[XMLGroupRange] = []
        self.areas: list[XMLArea] = []
        self.devices: list[DeviceInstance] = []
        self.language_code: str | None = None

        self.project_info: XMLProjectInformation
        self.functions: list[XMLFunction] = []

    def parse(self, language: str | None = None) -> KNXProject:
        """Parse ETS project."""
        self._load(language=language)
        self._sort()
        return self._transform()

    def _load(self, language: str | None) -> None:
        """Load XML files."""
        (
            knx_master_data,
            self.language_code,
        ) = KNXMasterLoader.load(
            knx_proj_contents=self.knx_proj_contents,
            knx_master_file=self.knx_proj_contents.root_path / "knx_master.xml",
            language=language,
        )
        (
            self.group_addresses,
            self.group_ranges,
            self.areas,
            self.devices,
            self.spaces,
            self.project_info,
            self.functions,
        ) = ProjectLoader.load(
            knx_proj_contents=self.knx_proj_contents,
            knx_master_data=knx_master_data,
        )

        products_dict: dict[str, Product] = {}
        hardware_application_map: HardwareToPrograms = {}
        for _products, _hardware_programs in [
            HardwareLoader.load(
                hardware_file=hardware_file,
                language_code=self.language_code,
            )
            for hardware_file in HardwareLoader.get_hardware_files(
                project_contents=self.knx_proj_contents
            )
        ]:
            products_dict.update(_products)
            hardware_application_map.update(_hardware_programs)

        for device in self.devices:
            device.manufacturer_name = knx_master_data.manufacturer_names.get(
                device.manufacturer, ""
            )

            try:
                product = products_dict[device.product_ref]
            except KeyError:
                _LOGGER.warning(
                    "Could not find hardware product for device %s with product_ref %s",
                    device.individual_address,
                    device.product_ref,
                )
                continue
            device.product_name = product.text
            device.hardware_name = product.hardware_name

            try:
                application_program_ref = hardware_application_map[
                    device.hardware_program_ref
                ]
            except KeyError:
                _LOGGER.warning(
                    "Could not find application_program_ref for device %s with hardware_program_ref %s",
                    device.individual_address,
                    device.hardware_program_ref,
                )
                continue
            device.application_program_ref = application_program_ref
            for com_object in device.com_object_instance_refs:
                com_object.resolve_com_object_ref_id(
                    application_program_ref, self.knx_proj_contents
                )
            for module_instance_argument in device.module_instance_arguments():
                module_instance_argument.complete_ref_id(application_program_ref)

        application_programs = (
            ApplicationProgramLoader.get_application_program_files_for_devices(
                project_root_path=self.knx_proj_contents.root_path,
                devices=self.devices,
            )
        )
        # update self.devices items with its inherited values from their application programs
        for application_program_file, devices in application_programs.items():
            ApplicationProgramLoader.load(
                application_program_path=application_program_file,
                devices=devices,
                language_code=self.language_code,
            )
        for device in self.devices:
            device.apply_module_instance_arguments()

    def _sort(self) -> None:
        """Sort loaded structures as XML content is sorted by creation time."""

        def recursive_sort_spaces(spaces: list[XMLSpace]) -> None:
            for _space in spaces:
                _space.devices.sort(
                    key=lambda ia: tuple(ia.split("."))  # area > line > device
                )
                recursive_sort_spaces(_space.spaces)

        recursive_sort_spaces(self.spaces)

        self.group_addresses.sort(key=attrgetter("raw_address"))

        def recursive_sort_group_ranges(group_ranges: list[XMLGroupRange]) -> None:
            for _grs in group_ranges:
                recursive_sort_group_ranges(_grs.group_ranges)
            group_ranges.sort(key=attrgetter("range_start"))

        recursive_sort_group_ranges(self.group_ranges)

        for area in self.areas:
            for line in area.lines:
                line.devices.sort(key=attrgetter("address"))
            area.lines.sort(key=attrgetter("address"))
        self.areas.sort(key=attrgetter("address"))

        self.devices.sort(key=attrgetter("area_address", "line_address", "address"))

    def _transform(self) -> KNXProject:
        """Convert XML Data to KNXProject structure."""
        _ga_id_to_address = {ga.identifier: ga.address for ga in self.group_addresses}

        communication_objects: dict[str, CommunicationObject] = {}
        devices_dict: dict[str, Device] = {}
        for device in self.devices:
            device_com_objects: list[str] = []
            for com_object in device.com_object_instance_refs:
                if not com_object.links:
                    continue
                group_address_links = [
                    valid_link
                    for link in com_object.links
                    if (valid_link := _ga_id_to_address.get(link))
                ]
                if not group_address_links:
                    # skip orphaned ComObjectInstanceRef pointing only to non-existent GroupAddress
                    # see https://github.com/XKNX/knx-frontend/issues/71
                    continue
                com_object_key = f"{device.individual_address}/{com_object.ref_id}"
                communication_objects[com_object_key] = CommunicationObject(
                    name=com_object.name,  # type: ignore[typeddict-item]
                    number=com_object.number,  # type: ignore[typeddict-item]
                    text=com_object.text,  # type: ignore[typeddict-item]
                    function_text=com_object.function_text,  # type: ignore[typeddict-item]
                    description=com_object.description or "",
                    device_address=device.individual_address,
                    channel=com_object.channel,
                    dpts=com_object.datapoint_types,
                    object_size=com_object.object_size,  # type: ignore[typeddict-item]
                    flags=Flags(
                        read=com_object.read_flag,  # type: ignore[typeddict-item]
                        write=com_object.write_flag,  # type: ignore[typeddict-item]
                        communication=com_object.communication_flag,  # type: ignore[typeddict-item]
                        update=com_object.update_flag,  # type: ignore[typeddict-item]
                        read_on_init=com_object.read_on_init_flag,  # type: ignore[typeddict-item]
                        transmit=com_object.transmit_flag,  # type: ignore[typeddict-item]
                    ),
                    group_address_links=group_address_links,
                )
                device_com_objects.append(com_object_key)

            devices_dict[device.individual_address] = Device(
                name=device.product_name,
                hardware_name=device.hardware_name,
                description=device.description,
                manufacturer_name=device.manufacturer_name,
                individual_address=device.individual_address,
                project_uid=device.project_uid,
                communication_object_ids=device_com_objects,
                channels={
                    channel.ref_id: Channel(
                        identifier=channel.ref_id, name=channel.name
                    )
                    for channel in device.channels
                },
            )

        topology_dict: dict[str, Area] = {}
        for area in self.areas:
            lines_dict: dict[str, Line] = {}
            for line in area.lines:
                devices_topology: list[str] = []
                for device in line.devices:
                    devices_topology.append(device.individual_address)
                lines_dict[str(line.address)] = Line(
                    name=line.name,
                    description=line.description,
                    devices=devices_topology,
                    medium_type=MEDIUM_TYPES.get(line.medium_type, "Unknown"),
                )
            topology_dict[str(area.address)] = Area(
                name=area.name, description=area.description, lines=lines_dict
            )

        group_address_dict: dict[str, GroupAddress] = {
            group_address.address: GroupAddress(
                name=group_address.name,
                identifier=group_address.identifier,
                raw_address=group_address.raw_address,
                address=group_address.address,
                project_uid=group_address.project_uid,
                dpt=group_address.dpt,
                data_secure=bool(group_address.data_secure_key),
                communication_object_ids=[
                    com_object_id
                    for com_object_id, com_object in communication_objects.items()
                    if group_address.address in com_object["group_address_links"]
                ],
                description=group_address.description,
                comment=html.unescape(rtf_to_text(group_address.comment)),
            )
            for group_address in self.group_addresses
        }

        group_range_dict: dict[str, GroupRange] = _recursive_convert_group_range(
            self.group_ranges, self.project_info.group_address_style
        )

        space_dict: dict[str, Space] = _recursive_convert_spaces(self.spaces)

        functions_dict: dict[str, Function] = {
            function.identifier: _convert_functions(function)
            for function in self.functions
        }

        info = ProjectInfo(
            project_id=self.project_info.project_id,
            name=self.project_info.name,
            last_modified=self.project_info.last_modified,
            group_address_style=self.project_info.group_address_style.value,
            guid=self.project_info.guid,
            created_by=self.project_info.created_by,
            schema_version=self.project_info.schema_version,
            tool_version=self.project_info.tool_version,
            xknxproject_version=__version__,
            language_code=self.language_code,
        )

        return KNXProject(
            info=info,
            communication_objects=communication_objects,
            topology=topology_dict,
            devices=devices_dict,
            group_addresses=group_address_dict,
            group_ranges=group_range_dict,
            locations=space_dict,
            functions=functions_dict,
        )



"""Combine the project data for more details inferred from linked objects."""
from __future__ import annotations

from typing import NamedTuple

from xknxproject.models import CommunicationObject, DPTType, KNXProject


class DPTTuple(NamedTuple):
    """A tuple representation of a DPT."""

    # used to have a hashable object to generate sets and do equality checks
    main: int
    sub: int | None

    def to_dpttype(self) -> DPTType:
        """Return a DPTType representation."""
        return DPTType(main=self.main, sub=self.sub)


def combine_project(project: KNXProject) -> KNXProject:
    """Combine the parsed project data for more details inferred from linked objects."""
    for comm_object in project["communication_objects"].values():
        if not comm_object["dpts"]:
            comm_object["dpts"] = _get_dpt_from_object_size(comm_object["object_size"])

    for group_address in project["group_addresses"].values():
        if not group_address["dpt"]:
            comm_objects = [
                project["communication_objects"][co_id]
                for co_id in group_address["communication_object_ids"]
            ]
            group_address["dpt"] = _get_dpt_from_comm_objects(comm_objects)
    return project


def _get_dpt_from_object_size(object_size: str) -> list[DPTType]:
    """Infer a DPTType from a CommunicationObjects object_size."""
    if object_size == "1 Bit":
        return [DPTType(main=1, sub=None)]
    if object_size == "2 Bit":
        # ignoring DPT 23.x which also has 2 bits
        return [DPTType(main=2, sub=None)]
    if object_size == "4 Bit":
        return [DPTType(main=3, sub=None)]
    return []


def _get_dpt_from_comm_objects(
    comm_objects: list[CommunicationObject],
) -> DPTType | None:
    """Infer a DPTType from a set of CommunicationObject DPTs."""
    dpts: set[DPTTuple] = {
        DPTTuple(**dpt) for co in comm_objects for dpt in co["dpts"] if dpt
    }
    if not dpts:
        return None
    if len(dpts) == 1:
        return next(iter(dpts)).to_dpttype()
    # if they all share the same main type, use a generic DPTType
    mains = {dpt.main for dpt in dpts}
    if len(mains) == 1:
        _main = next(iter(mains))
        return DPTType(main=_main, sub=None)
    return None



"""Constants for the library."""
from typing import Final

ETS_6_SCHEMA_VERSION: Final = 21
ETS_5_7_SCHEMA_VERSION: Final = 20
ETS_5_6_SCHEMA_VERSION: Final = 14
ETS_4_2_SCHEMA_VERSION: Final = 11


MAIN_DPT: Final = "DPT"
MAIN_AND_SUB_DPT: Final = "DPST"

ETS4_PRODUCT_LANGUAGES: Final = [
    "cs-CZ",
    "da-DK",
    "de-DE",
    "el-GR",
    "en-US",
    "es-ES",
    "fi-FI",
    "fr-FR",
    "hu-HU",
    "is-IS",
    "it-IT",
    "ja-JP",
    "nb-NO",
    "nl-NL",
    "pl-PL",
    "pt-PT",
    "ro-RO",
    "ru-RU",
    "sk-SK",
    "sl-SI",
    "sv-SE",
    "tr-TR",
    "uk-UA",
    "zh-CN",
]



"""Exceptions for the project parser library."""


class XknxProjectException(Exception):
    """Library base exception class."""


class InvalidPasswordException(XknxProjectException):
    """Invalid password exception."""


class ProjectNotFoundException(XknxProjectException):
    """Project files not found in archive."""


class UnexpectedFileContent(XknxProjectException):
    """Unexpected file content."""



"""Define output type for parsed KNX project."""
from __future__ import annotations

from typing import TypedDict


class DPTType(TypedDict):
    """DPT type dictionary."""

    main: int
    sub: int | None


class Flags(TypedDict):
    """Flags for the group addresses and KOs."""

    read: bool
    write: bool
    communication: bool
    transmit: bool
    update: bool
    read_on_init: bool


class CommunicationObject(TypedDict):
    """Communication object dictionary."""

    name: str
    number: int
    text: str
    function_text: str
    description: str
    device_address: str
    channel: str | None
    dpts: list[DPTType]
    object_size: str
    group_address_links: list[str]
    flags: Flags


class Device(TypedDict):
    """Devices dictionary."""

    name: str
    hardware_name: str
    description: str
    manufacturer_name: str
    individual_address: str
    project_uid: int | None
    communication_object_ids: list[str]
    channels: dict[str, Channel]  # id: Channel


class Channel(TypedDict):
    """Channel typed dict."""

    identifier: str
    name: str


class Line(TypedDict):
    """Line typed dict."""

    name: str
    medium_type: str
    description: str | None
    devices: list[str]


class Area(TypedDict):
    """Area typed dict."""

    name: str
    description: str | None
    lines: dict[str, Line]


class GroupAddress(TypedDict):
    """GroupAddress typed dict."""

    name: str
    identifier: str
    raw_address: int
    address: str
    project_uid: int | None
    dpt: DPTType | None
    data_secure: bool
    communication_object_ids: list[str]
    description: str
    comment: str


class GroupRange(TypedDict):
    """GroupRange holding the actual GAs but no children (e.g. 'middle' in THREELEVEL)."""

    name: str
    address_start: int
    address_end: int
    comment: str
    group_addresses: list[str]
    group_ranges: dict[str, GroupRange]


class Space(TypedDict):
    """Space typed dict."""

    type: str
    identifier: str
    name: str
    usage_id: str | None
    usage_text: str
    number: str
    description: str
    project_uid: int | None
    devices: list[str]
    spaces: dict[str, Space]
    functions: list[str]


class Function(TypedDict):
    """Function typed dict."""

    function_type: str
    group_addresses: dict[str, GroupAddressRef]
    identifier: str
    name: str
    project_uid: int | None
    space_id: str
    usage_text: str


class GroupAddressRef(TypedDict):
    """GroupAddressRef typed dict."""

    address: str
    name: str
    project_uid: int | None
    role: str


class ProjectInfo(TypedDict):
    """Information about the project."""

    project_id: str
    name: str
    last_modified: str | None
    group_address_style: str
    guid: str
    created_by: str
    schema_version: str
    tool_version: str
    xknxproject_version: str
    language_code: str | None


class KNXProject(TypedDict):
    """KNXProject typed dictionary."""

    info: ProjectInfo
    communication_objects: dict[str, CommunicationObject]
    devices: dict[str, Device]
    topology: dict[str, Area]
    locations: dict[str, Space]
    group_addresses: dict[str, GroupAddress]
    group_ranges: dict[str, GroupRange]
    functions: dict[str, Function]



"""Hardware Loader."""
from __future__ import annotations

from xml.etree import ElementTree
from zipfile import Path

from xknxproject.models import HardwareToPrograms, Product
from xknxproject.zip import KNXProjContents


class HardwareLoader:
    """Load hardware from KNX XML."""

    @staticmethod
    def load(
        hardware_file: Path,
        language_code: str | None,
    ) -> tuple[dict[str, Product], HardwareToPrograms]:
        """Load Hardware mappings."""
        product_dict: dict[str, Product] = {}
        hardware_programs: HardwareToPrograms = {}

        with hardware_file.open(mode="rb") as hardware_xml:
            tree = ElementTree.parse(hardware_xml)
            for hardware_node in tree.findall(
                ".//{*}Manufacturer/{*}Hardware/{*}Hardware"
            ):
                _products, _hardware_programs = HardwareLoader.parse_hardware_element(
                    hardware_node
                )
                product_dict |= _products
                hardware_programs |= _hardware_programs

            if language_code:
                for translation_element in tree.findall(
                    ".//{*}Manufacturer/{*}Languages"
                    f"/{{*}}Language[@Identifier='{language_code}']"
                    "/{*}TranslationUnit/{*}TranslationElement"
                ):
                    _ref_id = translation_element.get("RefId")
                    if _ref_id not in product_dict:
                        continue
                    HardwareLoader.apply_product_translation(
                        product_dict[_ref_id], translation_element
                    )

        return product_dict, hardware_programs

    @staticmethod
    def parse_hardware_element(
        hardware_node: ElementTree.Element,
    ) -> tuple[dict[str, Product], HardwareToPrograms]:
        """Parse hardware mapping."""
        product_dict: dict[str, Product] = {}
        hardware_programs: HardwareToPrograms = {}

        hardware_name: str = hardware_node.get("Name", "")
        for product_node in hardware_node.findall("{*}Products/{*}Product"):
            _product = HardwareLoader.parse_product_element(product_node)
            _product.hardware_name = hardware_name
            product_dict[_product.identifier] = _product

        for product_node in hardware_node.findall(
            "{*}Hardware2Programs/{*}Hardware2Program[@Id]/{*}ApplicationProgramRef[@RefId]/.."
        ):
            identifier, application_ref = HardwareLoader.parse_hardware2program_element(
                product_node
            )
            hardware_programs[identifier] = application_ref

        return product_dict, hardware_programs

    @staticmethod
    def parse_product_element(product_node: ElementTree.Element) -> Product:
        """Parse product mapping."""
        return Product(
            identifier=product_node.get("Id", ""),
            text=product_node.get("Text", ""),
        )

    @staticmethod
    def apply_product_translation(
        product: Product,
        translation_element_node: ElementTree.Element,
    ) -> None:
        """Apply translation to product."""
        if (
            text_node := translation_element_node.find(
                "{*}Translation[@AttributeName='Text']"
            )
        ) is not None:
            product.text = text_node.get("Text", "")

    @staticmethod
    def parse_hardware2program_element(
        hardware_to_program_node: ElementTree.Element,
    ) -> tuple[str, str]:
        """Parse hardware2program mapping."""
        identifier: str = hardware_to_program_node.get("Id", "")
        application_program_node = hardware_to_program_node.find(
            "{*}ApplicationProgramRef"
        )
        application_ref = application_program_node.get("RefId", "")  # type: ignore[union-attr]

        return identifier, application_ref

    @staticmethod
    def get_hardware_files(project_contents: KNXProjContents) -> list[Path]:
        """Get all manufactures Hardware.xml in given KNX ZIP file."""
        # M-*/Hardware.xml
        manufacturer_dirs = [
            child
            for child in project_contents.root_path.iterdir()
            if child.is_dir() and child.name.startswith("M-")
        ]
        return [
            xml_file
            for manufacturer in manufacturer_dirs
            if (xml_file := (manufacturer / "Hardware.xml")).exists()
        ]



"""Test parsing ETS projects."""

import pytest

from xknxproject import XKNXProj

from . import RESOURCES_PATH
from .conftest import assert_stub


@pytest.mark.parametrize(
    ("file_stem", "password", "language"),
    [
        ("xknx_test_project", "test", None),
        ("test_project-ets4", "test", "de-DE"),
        (
            "module-definition-test",
            None,
            "De",
        ),  # resolves to "de-DE" in parser for knx_master.xml
        (
            "testprojekt-ets6-functions",
            None,
            "De",
        ),  # resolves to "de-DE" in parser for knx_master.xml
        ("ets6_two_level", None, "de-DE"),
        ("ets6_free", None, "de-DE"),
    ],
)
def test_parse_project(file_stem: str, password: str, language: str):
    """Test parsing of various ETS projects (see pytest parameters)."""
    knxproj = XKNXProj(
        RESOURCES_PATH / f"{file_stem}.knxproj", password, language=language
    )
    project = knxproj.parse()
    assert_stub(project, f"{file_stem}.json")



"""Xknxproj models."""
# flake8: noqa
from .knxproject import (
    Area,
    Channel,
    CommunicationObject,
    Device,
    DPTType,
    Flags,
    Function,
    GroupAddress,
    GroupAddressRef,
    GroupRange,
    KNXProject,
    Line,
    ProjectInfo,
    Space,
)
from .models import (
    ChannelNode,
    ComObject,
    ComObjectInstanceRef,
    ComObjectRef,
    DeviceInstance,
    HardwareToPrograms,
    KNXMasterData,
    ModuleInstance,
    ModuleInstanceArgument,
    Product,
    TranslationsType,
    XMLArea,
    XMLFunction,
    XMLGroupAddress,
    XMLGroupAddressRef,
    XMLGroupRange,
    XMLLine,
    XMLProjectInformation,
    XMLSpace,
)
from .static import MEDIUM_TYPES, GroupAddressStyle, SpaceType



"""ETS Project Parser is a library to parse ETS project files."""
from __future__ import annotations

import logging
from pathlib import Path
import time

from xknxproject.__version__ import __version__
from xknxproject.combination import combine_project
from xknxproject.models import KNXProject
from xknxproject.xml import XMLParser
from xknxproject.zip.extractor import extract

_LOGGER = logging.getLogger("xknxproject.log")


class XKNXProj:
    """Class for parsing ETS project files."""

    def __init__(
        self,
        path: str | Path,
        password: str | None = None,
        language: str | None = None,
    ):
        """Initialize a KNXProjParser."""
        self.path = Path(path)
        self.password = password
        self.language = language

    def parse(self, combine: bool = True) -> KNXProject:
        """Parse the KNX project."""
        _LOGGER.info(
            'Xknxproject version %s parsing "%s" with%s password...',
            __version__,
            self.path,
            "" if self.password else "out",
        )
        _start = time.time()
        with extract(self.path, self.password) as knx_project_content:
            project = XMLParser(knx_project_content).parse(self.language)

        if combine:
            project = combine_project(project)

        _LOGGER.info("Parsing took %s seconds", time.time() - _start)
        _LOGGER.info(
            "Found %s group addresses, %s devices and %s used communication objects",
            len(project["group_addresses"]),
            len(project["devices"]),
            len(project["communication_objects"]),
        )
        return project



"""Test utilities."""
import pytest

from xknxproject.util import get_dpt_type, parse_dpt_types


@pytest.mark.parametrize(
    ("dpt_string", "expected"),
    [
        ("DPT-1", {"main": 1, "sub": None}),
        ("DPT-1 DPST-1-1", {"main": 1, "sub": None}),
        ("DPT-7 DPST-7-1", {"main": 7, "sub": None}),
        ("DPST-5-1", {"main": 5, "sub": 1}),
        ("DPT-1 DPT-5", {"main": 1, "sub": None}),
        ("DPT-14 DPST-14-1", {"main": 14, "sub": None}),
        ("DPST-6-10", {"main": 6, "sub": 10}),
        ("Wrong", None),
        ("DPT-Wrong", None),
        ("DPST-1-Wrong", None),
        ("DPST-5", None),
        ([], None),
        (None, None),
    ],
)
def test_get_dpt_type(dpt_string, expected):
    """Test parsing single DPT from ETS project."""
    assert get_dpt_type(dpt_string) == expected


@pytest.mark.parametrize(
    ("dpt_string", "expected"),
    [
        ("DPT-1", [{"main": 1, "sub": None}]),
        ("DPT-1 DPST-1-1", [{"main": 1, "sub": None}, {"main": 1, "sub": 1}]),
        ("DPT-7 DPST-7-1", [{"main": 7, "sub": None}, {"main": 7, "sub": 1}]),
        ("DPST-5-1", [{"main": 5, "sub": 1}]),
        ("DPT-1 DPT-5", [{"main": 1, "sub": None}, {"main": 5, "sub": None}]),
        ("DPT-14 DPST-14-1", [{"main": 14, "sub": None}, {"main": 14, "sub": 1}]),
        ("DPST-6-10", [{"main": 6, "sub": 10}]),
        ("Wrong", []),
        ("DPT-Wrong", []),
        ("DPST-1-Wrong", []),
        ("DPST-5", []),
        ([], []),
        (None, []),
    ],
)
def test_parse_dpt_types(dpt_string, expected):
    """Test parsing list of DPT from ETS project."""
    assert parse_dpt_types(dpt_string) == expected



"""XML utilities."""
from __future__ import annotations

import logging
from typing import overload

from xknxproject.const import MAIN_AND_SUB_DPT, MAIN_DPT
from xknxproject.models import DPTType

_LOGGER = logging.getLogger("xknxproject.log")


def get_dpt_type(dpt_string: str | None) -> DPTType | None:
    """Parse DPT type from the XML representation to main and sub types."""
    # GroupAddress tags should only support one single DPT.
    try:
        return parse_dpt_types(dpt_string)[0]
    except IndexError:
        return None


def parse_dpt_types(dpt_string: str | None) -> list[DPTType]:
    """Parse all DPTs from the XML representation to main and sub types."""
    if not dpt_string:
        return []

    supported_dpts: list[DPTType] = []
    # some applications have listed same DPT multiple times `DatapointType="DPST-1-1 DPST-1-1"`
    # so we use dict.fromkeys() (as set() doesn't preserve order)
    for _dpt in dict.fromkeys(dpt_string.split()):
        dpt_parts = _dpt.split("-")
        try:
            if MAIN_DPT == dpt_parts[0]:
                supported_dpts.append(
                    DPTType(
                        main=int(dpt_parts[1]),
                        sub=None,
                    )
                )
            if MAIN_AND_SUB_DPT == dpt_parts[0]:
                supported_dpts.append(
                    DPTType(
                        main=int(dpt_parts[1]),
                        sub=int(dpt_parts[2]),
                    )
                )
        except (IndexError, ValueError):
            _LOGGER.warning(
                'Could not parse DPTType from: "%s" in "%s"', _dpt, dpt_string
            )
    return supported_dpts


@overload
def parse_xml_flag(flag: str | None, default: bool) -> bool:
    ...


@overload
def parse_xml_flag(flag: str | None, default: None = None) -> bool | None:
    ...


def parse_xml_flag(flag: str | None, default: bool | None = None) -> bool | None:
    """Parse the XML flag to an optional boolean."""
    if flag is None:
        return default
    return flag == "Enabled"



"""Define internally used data structures."""
from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
import re

from xknxproject.models.knxproject import DPTType
from xknxproject.models.static import GroupAddressStyle, SpaceType
from xknxproject.zip import KNXProjContents

TranslationsType = dict[str, dict[str, str]]


class XMLGroupAddress:
    """Class that represents a group address."""

    def __init__(
        self,
        name: str,
        identifier: str,
        address: str,
        project_uid: int | None,
        description: str,
        dpt: DPTType | None,
        data_secure_key: str | None,
        comment: str,
        style: GroupAddressStyle,
    ):
        """Initialize a group address."""
        self.name = name
        self.identifier = identifier.split("_")[1]
        self.raw_address = int(address)
        self.project_uid = project_uid
        self.description = description
        self.dpt = dpt
        self.data_secure_key = data_secure_key  # Key as base64 encoded string or None
        self.comment = comment
        self.style = style
        self.address = XMLGroupAddress.str_address(self.raw_address, self.style)

    @staticmethod
    def str_address(raw_address: int, group_address_style: GroupAddressStyle) -> str:
        """Parse a given address and returns a string representation of it."""
        if group_address_style == GroupAddressStyle.FREE:
            return str(raw_address)
        main = (raw_address & 0b1111100000000000) >> 11
        if group_address_style == GroupAddressStyle.THREELEVEL:
            middle = (raw_address & 0b11100000000) >> 8
            sub = raw_address & 0b11111111
            return f"{main}/{middle}/{sub}"
        if group_address_style == GroupAddressStyle.TWOLEVEL:
            sub = raw_address & 0b11111111111
            return f"{main}/{sub}"
        raise ValueError(f"GroupAddressSyste '{group_address_style}' not supported!")

    def __repr__(self) -> str:
        """Return string representation."""
        return (
            f"{self.address} ({self.name}) - [DPT: {self.dpt}, ID: {self.identifier}]"
        )


@dataclass
class XMLGroupRange:
    """Class that represents a group range."""

    name: str
    range_start: int
    range_end: int
    group_addresses: list[int]
    group_ranges: list[XMLGroupRange]
    comment: str
    style: GroupAddressStyle

    def str_address(self) -> str:
        """Generate a string representation for the range."""
        if self.style == GroupAddressStyle.FREE:
            return f"{self.range_start}...{self.range_end}"
        if self.style == GroupAddressStyle.TWOLEVEL:
            return XMLGroupAddress.str_address(self.range_start, self.style).split("/")[
                0
            ]
        if self.style == GroupAddressStyle.THREELEVEL:
            start_address_token = XMLGroupAddress.str_address(
                self.range_start, self.style
            ).split("/")
            if (self.range_end - self.range_start) >= 2046:
                return start_address_token[0]
            return "/".join(start_address_token[0:2])
        raise ValueError(f"GroupAddressSyste '{self.style}' not supported!")


@dataclass
class XMLArea:
    """Class that represents a area."""

    address: int
    name: str
    description: str | None
    lines: list[XMLLine]


@dataclass
class XMLLine:
    """Class that represents a Line."""

    address: int
    description: str | None
    name: str
    medium_type: str
    devices: list[DeviceInstance]
    area: XMLArea


class DeviceInstance:
    """Class that represents a device instance."""

    def __init__(
        self,
        *,
        identifier: str,
        address: int,
        project_uid: int | None,
        name: str,
        description: str,
        last_modified: str,
        product_ref: str,
        hardware_program_ref: str,
        line: XMLLine,
        manufacturer: str,
        additional_addresses: list[str],
        channels: list[ChannelNode],
        com_object_instance_refs: list[ComObjectInstanceRef],
        module_instances: list[ModuleInstance],
        com_objects: list[ComObject] | None = None,
    ):
        """Initialize a Device Instance."""
        self.identifier = identifier
        self.address = address
        self.name = name
        self.description = description
        self.project_uid = project_uid
        self.last_modified = last_modified
        self.product_ref = product_ref
        self.hardware_program_ref = hardware_program_ref
        self.line = line
        self.area_address = line.area.address  # used for sorting
        self.line_address = line.address  # used for sorting
        self.manufacturer = manufacturer
        self.additional_addresses = additional_addresses
        self.channels: list[ChannelNode] = channels
        self.com_object_instance_refs = com_object_instance_refs
        self.module_instances = module_instances
        self.com_objects = com_objects or []
        self.application_program_ref: str | None = None

        self.individual_address = (
            f"{self.area_address}.{self.line_address}.{self.address}"
        )
        self.product_name: str = ""
        self.hardware_name: str = ""
        self.manufacturer_name: str = ""

    def add_additional_address(self, address: str) -> None:
        """Add an additional individual address."""
        self.additional_addresses.append(
            f"{self.line.area.address}/{self.line.address}/{address}"
        )

    def application_program_xml(self) -> str:
        """Obtain the file name to the application program XML."""
        return f"{self.manufacturer}/{self.application_program_ref}.xml"

    def module_instance_arguments(self) -> Iterator[ModuleInstanceArgument]:
        """Iterate ModuleInstance arguments."""
        for _module_instance in self.module_instances:
            yield from _module_instance.arguments

    def _complete_channel_placeholders(self) -> None:
        """Replace placeholders in channel names with module instance arguments."""
        for channel in self.channels:
            if not (
                channel.ref_id.startswith("MD-")  # only applicable if modules used
                and "{{" in channel.name  # placeholders are denoted "{{name}}"
            ):
                continue

            module_instance_ref = channel.ref_id.split("_CH")[0]
            module_instance = next(
                mi
                for mi in self.module_instances
                if mi.identifier == module_instance_ref
            )
            for argument in module_instance.arguments:
                channel.name = channel.name.replace(
                    f"{{{{{argument.name}}}}}", argument.value
                )

    def _add_com_object_instance_numbers(self) -> None:
        """Add module base object number to merged ComObjectInstanceRef."""
        for coir in self.com_object_instance_refs:
            if (
                coir.base_number_argument_ref is None
                or not coir.ref_id.startswith("MD-")
                or coir.number is None  # only for type safety
            ):
                continue
            _module_instance = next(
                mi
                for mi in self.module_instances
                if coir.ref_id.startswith(f"{mi.identifier}_")
            )
            coir.number += next(
                int(arg.value)
                for arg in _module_instance.arguments
                if arg.ref_id == coir.base_number_argument_ref
            )

    def apply_module_instance_arguments(self) -> None:
        """Apply module instance arguments."""
        self._complete_channel_placeholders()
        self._add_com_object_instance_numbers()


@dataclass
class ChannelNode:
    """Class that represents a Node with Type Channel."""

    ref_id: str
    name: str


@dataclass
class ModuleInstance:
    """Class that represents a ModuleInstance."""

    identifier: str
    ref_id: str
    arguments: list[ModuleInstanceArgument]


@dataclass
class ModuleInstanceArgument:
    """Class that represents a ModuleInstance Argument."""

    ref_id: str
    value: str
    name: str = ""  # resolved from application by `ref_id` ModuleDefs/ModuleDef/Arguments/Argument

    def complete_ref_id(self, application_program_ref: str) -> None:
        """Prepend the ref_id with the application program ref."""
        self.ref_id = f"{application_program_ref}_{self.ref_id}"


@dataclass
class ComObjectInstanceRef:
    """Class that represents a ComObjectInstanceRef instance."""

    identifier: str | None  # "Id" - xs:ID
    # "RefId" - knx:RELIDREF - required - points to a ComObjectRef Id
    # initially stripped by the devices application_program_ref
    ref_id: str
    text: str | None  # "Text"
    function_text: str | None  # "FunctionText"
    read_flag: bool | None  # "ReadFlag" - knx:Enable_t
    write_flag: bool | None  # "WriteFlag" - knx:Enable_t
    communication_flag: bool | None  # "CommunicationFlag" - knx:Enable_t
    transmit_flag: bool | None  # "TransmitFlag" - knx:Enable_t
    update_flag: bool | None  # "UpdateFlag" - knx:Enable_t
    read_on_init_flag: bool | None  # "ReadOnInitFlag" - knx:Enable_t
    datapoint_types: list[DPTType]  # "DataPointType" - knx:IDREFS
    description: str | None  # "Description" - language dependent
    channel: str | None  # "ChannelId" - knx:IDREFS
    links: list[str] | None  # "Links" - knx:RELIDREFS

    # resolved via Hardware.xml from the containing DeviceInstance
    com_object_ref_id: str | None = None

    # only available form ComObject and ComObjectRef
    name: str | None = None
    object_size: str | None = None
    # only available form ComObject
    base_number_argument_ref: str | None = None  # optional in ComObject
    number: int | None = None  # required in ComObject

    def resolve_com_object_ref_id(
        self, application_program_ref: str, knx_proj_contents: KNXProjContents
    ) -> None:
        """Prepend the ref_id with the application program ref."""
        # Remove module and ModuleInstance occurrence as they will not be in the application program directly
        ref_id = re.sub(r"(M-\d+?_MI-\d+?_)", "", self.ref_id)
        if knx_proj_contents.is_ets4_project():
            self.com_object_ref_id = ref_id
        else:
            self.com_object_ref_id = f"{application_program_ref}_{ref_id}"

    def merge_from_application(self, com_object: ComObject | ComObjectRef) -> None:
        """Fill missing information with information parsed from the application program."""
        if self.name is None:
            self.name = com_object.name
        if self.text is None:
            self.text = com_object.text
        if self.function_text is None:
            self.function_text = com_object.function_text
        if self.object_size is None:
            self.object_size = com_object.object_size
        if self.read_flag is None:
            self.read_flag = com_object.read_flag
        if self.write_flag is None:
            self.write_flag = com_object.write_flag
        if self.communication_flag is None:
            self.communication_flag = com_object.communication_flag
        if self.transmit_flag is None:
            self.transmit_flag = com_object.transmit_flag
        if self.update_flag is None:
            self.update_flag = com_object.update_flag
        if self.read_on_init_flag is None:
            self.read_on_init_flag = com_object.read_on_init_flag
        if not self.datapoint_types:
            self.datapoint_types = com_object.datapoint_types
        if isinstance(com_object, ComObject):
            self.number = com_object.number
            self.base_number_argument_ref = com_object.base_number_argument_ref


@dataclass
class ComObject:
    """Class that represents a ComObject instance."""

    __slots__ = (
        "identifier",
        "name",
        "text",
        "number",
        "function_text",
        "object_size",
        "read_flag",
        "write_flag",
        "communication_flag",
        "transmit_flag",
        "update_flag",
        "read_on_init_flag",
        "datapoint_types",
        "base_number_argument_ref",
    )

    # all items required in the XML
    identifier: str  # "Id" - xs:ID
    name: str  # "Name"
    text: str  # "Text" - language dependent
    number: int  # "Number" - xs:unsignedInt
    function_text: str  # "FunctionText" - language dependent
    object_size: str  # "ObjectSize" - knx:ComObjectSize_t
    read_flag: bool  # "ReadFlag" - knx:Enable_t
    write_flag: bool  # "WriteFlag" - knx:Enable_t
    communication_flag: bool  # "CommunicationFlag" - knx:Enable_t
    transmit_flag: bool  # "TransmitFlag" - knx:Enable_t
    update_flag: bool  # "UpdateFlag" - knx:Enable_t
    read_on_init_flag: bool  # "ReadOnInitFlag" - knx:Enable_t
    datapoint_types: list[DPTType]  # "DataPointType" - knx:IDREFS - optional
    # "BaseNumber" - knx:IDREF - optional - schema version >= 20
    # ModuleArgument identifier that holds value to add for
    # communication object number of ComObjectInstanceRef
    base_number_argument_ref: str | None


@dataclass
class ComObjectRef:
    """Class that represents a ComObjectRef instance."""

    __slots__ = (
        "identifier",
        "ref_id",
        "name",
        "text",
        "function_text",
        "object_size",
        "read_flag",
        "write_flag",
        "communication_flag",
        "transmit_flag",
        "update_flag",
        "read_on_init_flag",
        "datapoint_types",
    )

    identifier: str  # "Id" - xs:ID - required
    ref_id: str  # "RefId" - knx:IDREF - required - points to a ComObject Id
    name: str | None  # "Name"
    text: str | None  # "Text" - language dependent
    function_text: str | None  # "FunctionText" - language dependent
    object_size: str | None  # "ObjectSize" - knx:ComObjectSize_t
    read_flag: bool | None  # "ReadFlag" - knx:Enable_t
    write_flag: bool | None  # "WriteFlag" - knx:Enable_t
    communication_flag: bool | None  # "CommunicationFlag" - knx:Enable_t
    transmit_flag: bool | None  # "TransmitFlag" - knx:Enable_t
    update_flag: bool | None  # "UpdateFlag" - knx:Enable_t
    read_on_init_flag: bool | None  # "ReadOnInitFlag" - knx:Enable_t
    datapoint_types: list[DPTType]  # "DataPointType" - knx:IDREFS


@dataclass
class KNXMasterData:
    """KNX Master data needed for parsing other project files."""

    function_type_names: dict[str, str]
    manufacturer_names: dict[str, str]
    space_usage_mapping: dict[str, str]
    translations: TranslationsType

    def _get_translation_item(
        self, ref_id: str, attribute_name: str = "Text"
    ) -> str | None:
        """Get translation item from the translations dict."""
        if self.translations:
            try:
                return self.translations[ref_id][attribute_name]
            except KeyError:
                return None
        return None

    def get_function_type_name(self, function_type_id: str) -> str:
        """Get space usage name from space usage id."""
        if translated := self._get_translation_item(function_type_id):
            return translated
        return self.function_type_names.get(function_type_id, "")

    def get_space_usage_name(self, space_usage_id: str) -> str:
        """Get space usage name from space usage id."""
        if translated := self._get_translation_item(space_usage_id):
            return translated
        return self.space_usage_mapping.get(space_usage_id, "")


@dataclass
class XMLSpace:
    """A space in the location XML."""

    identifier: str
    name: str
    space_type: SpaceType
    usage_id: str | None  # SU-<int> resolved from knx_master.xml (with translation)
    usage_text: str  # default to "" - translated
    number: str  # default to ""
    description: str  # default to ""
    project_uid: int | None
    spaces: list[XMLSpace]
    devices: list[str]  # [DeviceInstance.individual_address]
    functions: list[str]


@dataclass
class XMLFunction:
    """A functions in the space XML."""

    function_type: str
    group_addresses: list[XMLGroupAddressRef]
    identifier: str
    name: str
    project_uid: int | None
    space_id: str
    usage_text: str


@dataclass
class XMLGroupAddressRef:
    """A GroupAddressRef in the functions XML."""

    address: str
    identifier: str
    name: str
    project_uid: int | None
    ref_id: str
    role: str


@dataclass
class Product:
    """Model a Product instance."""

    identifier: str
    text: str
    hardware_name: str = ""


HardwareToPrograms = dict[str, str]


@dataclass
class XMLProjectInformation:
    """Model a ProjectInformation instance."""

    # ProjectInformation tag is not required in XSD, thus everything is optional
    project_id: str = ""
    name: str = ""
    last_modified: str | None = None
    group_address_style: GroupAddressStyle = GroupAddressStyle.THREELEVEL
    guid: str = ""
    created_by: str = ""
    schema_version: str = ""
    tool_version: str = ""



"""XML loader for xknxproj files."""
from .application_program_loader import ApplicationProgramLoader
from .hardware_loader import HardwareLoader
from .knx_master_loader import KNXMasterLoader
from .project_loader import ProjectLoader

__all__ = [
    "ApplicationProgramLoader",
    "HardwareLoader",
    "KNXMasterLoader",
    "ProjectLoader",
]



"""Application Program Loader."""
from __future__ import annotations

from collections.abc import Iterator
import logging
from typing import Any
from xml.etree import ElementTree
from zipfile import Path

from xknxproject.models import ComObject, ComObjectRef, DeviceInstance
from xknxproject.util import parse_dpt_types, parse_xml_flag

_LOGGER = logging.getLogger("xknxproject.log")


class ApplicationProgramLoader:
    """Load the application program from KNX XML."""

    @staticmethod
    def load(
        application_program_path: Path,
        devices: list[DeviceInstance],
        language_code: str | None,
    ) -> None:  # tuple[dict[str, ComObjectRef], dict[str, ComObject]]:
        """Load Hardware mappings and assign to devices."""
        com_object_instance_refs = [
            instance_ref
            for device in devices
            for instance_ref in device.com_object_instance_refs
        ]
        used_com_object_ref_ids = {
            instance_ref.com_object_ref_id
            for instance_ref in com_object_instance_refs
            if instance_ref.com_object_ref_id is not None
        }
        com_object_refs: dict[str, ComObjectRef] = {}  # {Id: ComObjectRef}
        com_objects: dict[str, ComObject] = {}  # {Id: ComObject}

        used_module_arguments = {
            attribute.ref_id: ""
            for device in devices
            for attribute in device.module_instance_arguments()
        }

        with application_program_path.open(mode="rb") as application_xml:
            tree_iterator = ElementTree.iterparse(application_xml, events=("start",))
            for _, elem in tree_iterator:
                if elem.tag.endswith("ComObject"):
                    # we take all since we don't know which are referenced to yet
                    identifier = elem.attrib.get("Id")
                    com_objects[identifier] = ApplicationProgramLoader.parse_com_object(
                        elem, identifier
                    )
                elif elem.tag.endswith("ComObjectRef"):
                    if (_id := elem.attrib.get("Id")) in used_com_object_ref_ids:
                        com_object_refs[
                            _id
                        ] = ApplicationProgramLoader.parse_com_object_ref(elem, _id)
                    elem.clear()
                elif elem.tag.endswith("Argument"):  # ModuleDefs/ModuleDef/Arguments/
                    if (_id := elem.attrib.get("Id")) in used_module_arguments:
                        used_module_arguments[_id] = elem.attrib.get("Name")
                    elem.clear()
                elif elem.tag.endswith("Languages"):
                    elem.clear()
                    # hold iterator for optional translation parsing
                    break
                elem.clear()

            if language_code is not None:
                ApplicationProgramLoader.parse_translations(
                    tree_iterator=tree_iterator,
                    com_objects=com_objects,
                    com_object_refs=com_object_refs,
                    used_com_object_ref_ids=used_com_object_ref_ids,
                    language_code=language_code,
                )

            for device in devices:
                for attribute in device.module_instance_arguments():
                    attribute.name = used_module_arguments[attribute.ref_id]

            for com_instance in com_object_instance_refs:
                if com_instance.com_object_ref_id is None:
                    _LOGGER.warning(
                        "ComObjectInstanceRef %s has no ComObjectRefId",
                        com_instance.identifier,
                    )
                    continue
                _com_object_ref = com_object_refs[com_instance.com_object_ref_id]
                com_instance.merge_from_application(_com_object_ref)
                com_instance.merge_from_application(com_objects[_com_object_ref.ref_id])

    @staticmethod
    def parse_translations(
        tree_iterator: Iterator[tuple[str, Any]],
        com_objects: dict[str, ComObject],
        com_object_refs: dict[str, ComObjectRef],
        used_com_object_ref_ids: set[str],
        language_code: str,
    ) -> None:
        """Parse translations. Replace translated text in com_objects and com_object_refs."""
        used_com_object_ids = {
            com_object_ref.ref_id for com_object_ref in com_object_refs.values()
        }
        used_translation_ids = used_com_object_ids | used_com_object_ref_ids
        in_language = False
        in_translation_ref: str | None = None  # TranslationElement RefId
        # translation_map: {TranslationElement RefId: {AttributeName: Text}}
        translation_map: dict[str, dict[str, str]] = {}
        for _, elem in tree_iterator:
            if elem.tag.endswith("Language"):
                if in_language:
                    # Already found the language we are looking for.
                    # We don't need anything after that tag (there isn't much anyway)
                    elem.clear()
                    break
                in_language = elem.get("Identifier") == language_code
            elif in_language and elem.tag.endswith("TranslationElement"):
                ref_id = elem.get("RefId")
                in_translation_ref = ref_id if ref_id in used_translation_ids else None
            elif (
                in_language
                and in_translation_ref is not None
                and elem.tag.endswith("Translation")
            ):
                translation_map.setdefault(in_translation_ref, {})[
                    elem.get("AttributeName")
                ] = elem.get("Text")
            elem.clear()

        ApplicationProgramLoader.apply_translations(com_object_refs, translation_map)
        ApplicationProgramLoader.apply_translations(com_objects, translation_map)

    @staticmethod
    def parse_com_object(
        elem: ElementTree.Element,
        identifier: str,
    ) -> ComObject:
        """Parse ComObject tag."""
        return ComObject(
            identifier=identifier,
            name=elem.get("Name"),  # type: ignore[arg-type]
            text=elem.get("Text"),  # type: ignore[arg-type]
            number=int(elem.get("Number", 0)),
            function_text=elem.get("FunctionText"),  # type: ignore[arg-type]
            object_size=elem.get("ObjectSize"),  # type: ignore[arg-type]
            read_flag=parse_xml_flag(elem.get("ReadFlag"), False),
            write_flag=parse_xml_flag(elem.get("WriteFlag"), False),
            communication_flag=parse_xml_flag(elem.get("CommunicationFlag"), False),
            transmit_flag=parse_xml_flag(elem.get("TransmitFlag"), False),
            update_flag=parse_xml_flag(elem.get("UpdateFlag"), False),
            read_on_init_flag=parse_xml_flag(elem.get("ReadOnInitFlag"), False),
            datapoint_types=parse_dpt_types(elem.get("DatapointType")),
            base_number_argument_ref=elem.get("BaseNumber"),
        )

    @staticmethod
    def parse_com_object_ref(
        elem: ElementTree.Element,
        identifier: str,
    ) -> ComObjectRef:
        """Parse ComObjectRef tag."""
        return ComObjectRef(
            identifier=identifier,
            ref_id=elem.get("RefId"),  # type: ignore[arg-type]
            name=elem.get("Name"),
            text=elem.get("Text"),
            function_text=elem.get("FunctionText"),
            object_size=elem.get("ObjectSize"),
            read_flag=parse_xml_flag(elem.get("ReadFlag")),
            write_flag=parse_xml_flag(elem.get("WriteFlag")),
            communication_flag=parse_xml_flag(elem.get("CommunicationFlag")),
            transmit_flag=parse_xml_flag(elem.get("TransmitFlag")),
            update_flag=parse_xml_flag(elem.get("UpdateFlag")),
            read_on_init_flag=parse_xml_flag(elem.get("ReadOnInitFlag")),
            datapoint_types=parse_dpt_types(elem.get("DatapointType")),
        )

    @staticmethod
    def apply_translations(
        com_objects: dict[str, ComObject] | dict[str, ComObjectRef],
        translation_map: dict[str, dict[str, str]],
    ) -> None:
        """Apply translations to ComObjects and ComObjectRefs."""
        for identifier in com_objects.keys() & translation_map.keys():
            translation = translation_map[identifier]
            com_object = com_objects[identifier]
            if _text := translation.get("Text"):
                com_object.text = _text
            if _function_text := translation.get("FunctionText"):
                com_object.function_text = _function_text

    @staticmethod
    def get_application_program_files_for_devices(
        project_root_path: Path,
        devices: list[DeviceInstance],
    ) -> dict[Path, list[DeviceInstance]]:
        """Do not load the same application program multiple times."""
        _result: dict[str, list[DeviceInstance]] = {}
        for device in devices:
            if device.application_program_ref:
                # zipfile.Path hashes are not equal, therefore we use str to create the struct
                xml_file_name = device.application_program_xml()
                _result.setdefault(xml_file_name, []).append(device)

        return {
            (project_root_path / xml_file): devices
            for xml_file, devices in _result.items()
        }
